/*
 * Copyright (c) 2000-2012 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** common code for KVH 1000 and 5000 gyros
 **/

#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <portLib.h>
#include <errnoLib.h>
#include <taskLib.h>
#include <commonStructLib.h>
#include <h2mathLib.h>

#include <gyroLib/gyro.h>
#include <gyroLib/gyroTypes.h>

#include "gyroPrivate.h"

/*----------------------------------------------------------------------
 * gyroInit -
 */
struct GYRO_DATA *
gyroInit(int type, const char *device,
    double latitude, double woffset)
{
	switch (type) {
	case GYRO_KVH_ECORE_1000:
		return gyroInitKVH1000(device, latitude, woffset);
	case GYRO_KVH_DSP_5000:
		return gyroInitKVH5000(device, latitude, woffset, -1);
	case GYRO_KVH_DSP_3000:
		return gyroInitKVH5000(device, latitude, woffset, -1);
	default:
		errnoSet(S_gyroLib_DATA_NOT_VALID);
		return NULL;
	}
}

/* ----------------------------------------------------------------------
 *
 * gyroEnd -     End connection to serial port
 *
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL
gyroEnd(struct GYRO_DATA *gyroData)
{
	if (gyroData == NULL)
		return TRUE;

	if (gyroData->close(gyroData) == FALSE) {
		return FALSE;
	}
	return TRUE;
}

/* ----------------------------------------------------------------------
 *
 * gyroReadAngle -
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL
gyroReadAngle(struct GYRO_DATA *gyroData, double *theta)
{
	if (gyroData == NULL) {
		errnoSet(S_gyroLib_NOT_INITIALIZED);
		return FALSE;
	}

	commonStructTake(gyroData);
	*theta = gyroData->theta;
	commonStructGive(gyroData);
	return TRUE;
}

/*
 * gyroReadW ----------------------------------------------------------------
 *
 * WARNING : return TRUE or FALSE
 */

BOOL
gyroReadW(struct GYRO_DATA *gyroData, double *w)
{
	if (gyroData == NULL) {
		errnoSet(S_gyroLib_NOT_INITIALIZED);
		return FALSE;
	}

	commonStructTake(gyroData);
	*w = gyroData->w;
	commonStructGive(gyroData);

	return TRUE;
}

BOOL
gyroRead(struct GYRO_DATA *gyroData, double *theta, double *w, uint32_t *date)
{

	if (gyroData == NULL) {
		errnoSet(S_gyroLib_NOT_INITIALIZED);
		return FALSE;
	}

	commonStructTake(gyroData);
	*w = gyroData->w;
	*theta = gyroData->theta;
	*date = gyroData->date;
	commonStructGive(gyroData);

	return TRUE;
}
/*
 * gyroUpdateWOffset -----------------------------------------------------
 *
 * Computes the w offset (theoritically depends on temperature)
 * Controls also the period
 *
 *         WARNING: the robot must be stopped
 *
 * It returns when the computation is over
 */


BOOL
gyroUpdateWOffset(struct GYRO_DATA *gyroData, int nbMeasures)
{
	double wOffset;
	int nbMeasuresOffset;

	/* not init */
	if (gyroData == NULL) {
		errnoSet(S_gyroLib_NOT_INITIALIZED);
		return FALSE;
	}

	/* thus need to wait for waitDuration seconds */
	printf ("  gyro period = %u ms (NTICKS_PER_SEC = %d)\n",
	    gyroData->period,  NTICKS_PER_SEC);
	printf ("  gyroWOffset: needs %d seconds (%d measures) with robot STOPPED !!\n",
	    (int)(nbMeasures*gyroData->period/1000.0),
	    nbMeasures);


	/* init mesures */
	commonStructTake(gyroData);
	gyroData->wOffset = 0.0;
	gyroData->nbMeasuresOffset = nbMeasures;
	commonStructGive(gyroData);

	/* mesures */
	nbMeasuresOffset = nbMeasures;
	while (nbMeasuresOffset) {
		/* we do not take the common-struct in purpose to do not disturb
		   the main task. this value is just for display */
		nbMeasuresOffset =  gyroData->nbMeasuresOffset;
		printf ("%d ", nbMeasuresOffset);
		fflush(stdout);
		taskDelay(NTICKS_PER_SEC);
	}

	/* print result */
	commonStructTake(gyroData);
	wOffset = gyroData->wOffset;
	commonStructGive(gyroData);
	printf ("\ngyroUpdateWOffset: offset %.3f deg/s\n", wOffset);

	return TRUE;
}

/*----------------------------------------------------------------------
 *
 * gyroReadWOffset
 */
double
gyroReadWOffset(struct GYRO_DATA *gyro)
{
	double result;

	commonStructTake(gyro);
	result = gyro->wOffset;
	commonStructGive(gyro);
	return result;
}

/* ----------------------------------------------------------------------
 *
 * gyroResetAngle - Reset the integrated angle to the value parameter
 *
 * Returns :            FALSE if failed or TRUE.
 */

BOOL
gyroResetAngle(struct GYRO_DATA *gyroData, double angle)
{
	if (gyroData == NULL) {
		errnoSet(S_gyroLib_NOT_INITIALIZED);
		return FALSE;
	}
	commonStructTake(gyroData);
	gyroData->theta = angleLimit(angle);
	commonStructGive(gyroData);
	return TRUE;
}
/*----------------------------------------------------------------------
 *
 * gyroEarthSpeed - return the local projection of the earth rotation
 *                  speed in rad/s
 *
 * 86164 is the duration in seconds of a sideral day (23h56'4")
 */
double
gyroEarthSpeed(double latitude)
{
	return 2*M_PI/86164*sin(latitude);
}

/*----------------------------------------------------------------------*/

/*
 * gyroIntegrate - common code to compute the angle from velocity
 *
 * must be called with commonStruct locked
 */
void
gyroIntegrate(struct GYRO_DATA *gyro, double w)
{
	static int totMeasuresOffset = 0;

	/*
	 * offset correction
	 */
	if (gyro->nbMeasuresOffset) {
		/* calibration on: estimation of wOffset ongoing */
		if (totMeasuresOffset == 0)
			totMeasuresOffset = gyro->nbMeasuresOffset;
		gyro->wOffset += w/totMeasuresOffset;
		gyro->nbMeasuresOffset--;
	} else {
		/* calibration off: use wOffset to correct w */
		gyro->w = w - gyro->wOffset;
	}
	/*
	 * compute angle
	 */
	if (gyro->prevDate)
		gyro->period = gyro->date - gyro->prevDate;
	else {
		gyro->period = 0;
	}
	gyro->prevDate = gyro->date;
	gyro->theta += gyro->w * gyro->period/1000.0;
	gyro->theta = angleLimit(gyro->theta);
}
